#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 21 10:20:51 2021

@author: fsvbach
"""

from scipy.stats import special_ortho_group, wishart
from scipy.linalg import eigh
import numpy as np
import pandas as pd
import time

def RotationMatrix(degree):
    degree = np.deg2rad(degree)
    return np.array([[np.cos(degree),-np.sin(degree)],
                     [np.sin(degree), np.cos(degree)]]) 

def MirrorMatrix(vector):
    x,y = vector
    return np.array([[x**2-y**2,2*x*y],
                     [2*x*y, y**2-x**2]]) / np.linalg.norm(vector)**2

class CovarianceMatrix:
    def __init__(self, P=RotationMatrix(45), s=np.array([1,4])):
        '''
        stores eigen-decomposition of covariance matrix
        ----------
        P : np.ndarray
            orthogonal matrix.
        s : np.array
            eigenvalues in array.
        '''       
        self.P = P
        self.s = s
        
    def diagonalize(self):
        self.s  = np.diag(self.array())
        self.P  = np.eye(len(self.s))
        
    def normalize(self):
        A = self.array()
        diagonal = np.sqrt(np.diag(A)) + 0.00000001
        A = (A/diagonal).T / diagonal
        self.s, self.P = eigh(A)
        self.s[np.where(self.s<0)]=0
            
    def array(self):
        return self.P@np.diag(self.s)@self.P.T
    
    def sqrt(self):
        return self.P@np.diag(np.sqrt(self.s))@self.P.T
    
    def shape(self, std=1):
        assert len(self.s) == 2
        angle         = np.degrees(np.arctan2(*self.P[:,0][::-1]))
        width, height = np.sqrt(self.s)*std*2
        return width, height, angle 

def arr2cov(array):
    s, P = eigh(array)
    s[np.where(s<0)]=0
    assert np.all(s>=0)
    return CovarianceMatrix(P, s)  

class WishartDistribution:
    def __init__(self, nu=2, scale=CovarianceMatrix()):
        self.nu = nu
        self.scale  = scale
        
    def shape(self, std=1):
        return self.scale.shape(std=std)
        
class GaussianDistribution:
    def __init__(self, mean=np.array([1,0]), cov=CovarianceMatrix()):
        self.mean = mean
        self.cov  = cov

    def estimate(self, data):
        N = len(data)
        assert N > 1
        self.mean = np.mean(data, axis=0)
        self.cov  = arr2cov((data-self.mean).T @ (data-self.mean) / (N-1))
        return GaussianDistribution(self.mean, self.cov)
    
    def shape(self, std=1):
        width, height, angle = self.cov.shape(std=std)
        return self.mean, width, height, angle

    def samples(self, size=20, seed=None):
        return np.random.default_rng(seed=seed).multivariate_normal(mean = self.mean, 
                                                  cov  = self.cov.array(),
                                                  size = size) 

def Dataset2Gaussians(dataset, diagonal=False, normalize=False):
    Gaussians = []
    names    = []
    for name, data in dataset.groupby(level=0):
        G = GaussianDistribution()
        G.estimate(data.values)
        if diagonal:
            G.cov.diagonalize()
        elif normalize:
            G.cov.normalize()
        names.append(name)
        Gaussians.append(G)
    return pd.Series(Gaussians, index=names)

class RandomGenerator:
    def __init__(self, seed=None):
        self.generator = np.random.default_rng(seed=seed)
    
    def RandomSeed(self):
        return self.UniformInteger(upper=10000)[0]
    
    def UniformInteger(self, lower=0, upper=10, size=1):
        return self.generator.integers(lower, upper, size)
    
    def UniformVector(self, dim=2, upper=1, size=1):
        samples = self.generator.random((size, dim)) - 0.5 
        return 2*samples*upper
    
    def OrthogonalMatrix(self, dim=2, size=1):
        return special_ortho_group.rvs(dim=dim, random_state=self.RandomSeed(), size=size)
    
    def UniformCovariance(self, dim=2, maxstd=1):
        P = self.OrthogonalMatrix(dim=dim)
        s = (self.generator.random(size=dim) * maxstd)**2
        return CovarianceMatrix(P,s)
        
    def GaussianSamples(self, Gaussian, size=1):
        return self.generator.multivariate_normal(mean = Gaussian.mean, 
                                                  cov  = Gaussian.cov.array(),
                                                  size = size)
    def WishartSamples(self, Wishart, size=1):
        arrays = wishart.rvs(Wishart.nu, Wishart.scale.array(), random_state=self.RandomSeed(), size=size)
        return [arr2cov(arr) for arr in arrays]

class Timer:
    def __init__(self, name, dec=3, output=True):
        self.name      = name
        self.dec       = 3
        self.output    = output
        self.start     = self.time()
        self.last_time = self.start
        self.date      = self.date()
        self.log       = [f"Started '{name}' at {self.date}\n"]
  
    def date(self):
        t = time.localtime()
        return time.strftime("%m-%d-%H-%M-%S", t)
    
    def time(self):
        return round(time.perf_counter(), self.dec)
    
    def total_time(self):
        return round(self.time() - self.start, self.dec)
    
    def add(self, infomsg):
        time = self.time() - self.last_time
        msg  = f'{infomsg} in {round(time, self.dec)}s. (Total: {self.total_time()}s)\n'
        self.last_time = self.time()
        self.log.append(msg)
        print(msg)
    
    def result(self, result):
        self.log.append(result)
        print(result)

    def finish(self, location):
        infomsg = f"Succesfully finished '{self.name}' in {self.total_time()}s"
        print(infomsg)
        
        if self.output:
            for msg in reversed(self.log):
                infomsg = msg + '\n' + infomsg
            file = open(location, "a")
            file.write(infomsg)
            file.write(f"\n\n{'-'*75}\n")
            file.close() 
        
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 14 10:04:54 2021

@author: fsvbach
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse, FancyArrow
from matplotlib.legend_handler import HandlerPatch

def HandlerArrow(legend, orig_handle,
                      xdescent, ydescent,
                      width, height, fontsize):
    p = FancyArrow(0, 0.5*height, width, 0, length_includes_head=True, head_width=0.75*height )
    return p

class HandlerEllipseRotation(HandlerPatch):
    def create_artists(self, legend, orig_handle,
                       ydescent, xdescent, height, width, fontsize, trans):
        center = 0.5 * height, 0.5 * width
        p = Ellipse(xy=center, width=orig_handle.width,
                             height=orig_handle.height,
                             angle=orig_handle.angle, linewidth=.5)
        self.update_prop(p, orig_handle, legend)
        p.set_transform(trans)
        return [p]
    
class HandlerEllipse(HandlerPatch):
    def create_artists(self, legend, orig_handle,
                       xdescent, ydescent, width, height, fontsize, trans):
        center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
        p = Ellipse(xy=center, width=width + xdescent,
                             height=height + ydescent)
        self.update_prop(p, orig_handle, legend)
        p.set_transform(trans)
        return [p]
        
def plotMixture(mixture, std=2, ax=None):
    fig = None
    if not ax:
        fig, ax = plt.subplots(figsize=(7,7))
    
    # covlabel = rf'{std}-$\sigma$ class covariance'
    covlabel = 'Classes'
    dataset  = mixture.data.groupby(level=0).mean().sample(frac=1, random_state=43)
    dataset.index = dataset.index.to_series().map(mixture.labeldict())
    

    if std > 0:
        for Gaussian in mixture.ClassGaussians:
            # plotting black class covariances
            mean, width, height, angle = Gaussian.shape(std=std)
            ell = Ellipse(xy=mean, width=width, height=height, angle=angle, 
                          edgecolor='black', facecolor='none', 
                          linewidth=0.5, linestyle='--', 
                          label=covlabel, zorder=6)
            ax.add_patch(ell)
            covlabel=None
    
    # plotting colourful datapoints
    xmeans, ymeans = dataset.values.T
    ax.scatter(xmeans, ymeans, s=1, c=dataset.index, label='Unit means', zorder=3)

    # plotting grey samples
    xsample, ysample = mixture.data.values.T
    ax.scatter(xsample, ysample, s=0.5, linewidths=0, c='grey', label='Individual samples', zorder=2)
    
    handles = []
    labels  = []
    for i, Wishart in enumerate(mixture.ClassWisharts):
        # adding data covariances to 2nd legend 
        width, height, angle = Wishart.shape(std=2)
        ell = Ellipse(xy=(0,0), width=width, height=height, angle=angle, 
                      edgecolor="C"+str(i), facecolor='none', 
                      linewidth=0.5, 
                      label='class '+str(i+1))
        handles.append(ell)
        labels.append('Class ' +str(i))
        # ax.add_artist(ell)

    
    # storing 1st legend 
    leg1 = ax.legend(handler_map={Ellipse: HandlerEllipse()}, handletextpad=1,
                     loc='upper right', title="Hierarchical Structure",
                     facecolor='white', scatterpoints=4, framealpha=1)    
    
    leg1.legendHandles[2]._sizes = [2.5]
    
    # adding legends
    leg2 = ax.legend(handles, labels, handler_map={Ellipse: HandlerEllipseRotation()},
                    title="Wishart Scales", loc=("lower left"), ncol=int(np.ceil(mixture.K/2)), facecolor='white',
                    labelspacing=0.7, columnspacing=1, handleheight=1, handlelength=1)
    leg2.get_frame().set_linewidth(.5)
    leg1.get_frame().set_linewidth(.5)
    ax.add_artist(leg1)
    
    # add title
    ax.set_aspect('equal')
    ax.set_title(mixture.info)

    return fig

def plotGaussian(Gaussian, size=20, STDS=[1,2,3], color='black', lw=1, r=1, ax=None):
    fig = None
    if not ax:
        fig, ax = plt.subplots(figsize=(5,5))
        
    for i in STDS:
        mean, width, height, angle = Gaussian.shape(std=i)
        ell = Ellipse(xy=mean, width=width, height=height, angle=angle, 
                      edgecolor=color, facecolor='none', 
                      linewidth=lw, linestyle='--')
        ax.add_patch(ell)
        
    if size:
        samples = Gaussian.samples(size, seed=13)
        x,y = samples.T
        ax.scatter(x,y, color=color, s=r)
        
    return fig

def embedScatter(embedding, labeldict=None, sizes=1, title='', ax=None):
    fig = None
    if not ax:
        fig, ax = plt.subplots(figsize=(5,5))
        
    if labeldict:
        embedding['labels'] = embedding.index.to_series().map(labeldict)
    else:
        embedding['labels'] = 'Units'
    embedding['sizes']  = sizes
    
    for label, data in embedding.groupby('labels'):
        ax.scatter(data['x'], data['y'], s=data['sizes'], label=label)
        
    ax.set_xticks([], minor=[])
    ax.set_yticks([], minor=[])
    ax.set_title(title)
    ax.legend()
    
    return fig
    
    
naming = {0: 'Euclidean', 0.5: 'Wasserstein', 1: 'Covariance'} 

def plotMatrices(matrices, titles=None):
    n = len(matrices)
    if not titles:
        titles = ['' for i in range(n)]
        
    fig, axes = plt.subplots(ncols=n, figsize=(6*n,5))
    for matrix, title, ax in zip(matrices, titles, axes):
        m = ax.imshow(matrix, cmap='Greens')
        ax.set_title(title)
        plt.colorbar(m, ax=ax)
    return fig



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  7 10:57:05 2021

@author: fsvbach
"""

import pandas as pd
import numpy as np
import scipy.sparse as sp

from scipy.optimize import linprog
from .utils import arr2cov, Timer

def EuclideanDistance(A,B):
    N1 = np.linalg.norm(A, ord=2, axis=1).reshape(-1,1)**2 
    N2 = np.linalg.norm(B, ord=2, axis=1).reshape(1,-1)**2 
    N3 = -2 * np.inner(A,B)
    D  = N1 + N2 + N3
    D[np.where(D<0)] = 0
    return np.sqrt(D)
    
def ConstraintMatrix(n,m):
    N = np.repeat(np.identity(n,dtype=int), m, axis=1)
    M = np.hstack([np.identity(m, dtype=int)]*n)
    return np.vstack([N,M])

def SparseConstraint(n,m):
    row  = np.repeat(np.arange(n), m)
    col  = np.arange(n*m) 
    data = np.ones(n*m, dtype=int)
    N = sp.csr_matrix((data, (row, col)), 
                   shape=(n, n*m))
    M = sp.hstack([sp.dia_matrix((np.ones(m, dtype=int), 0), shape=(m,m))]*n)
    return sp.vstack([N,M], format='csr')

def PairwiseWassersteinDistance(U, V, p=2):
    uniqueRowsU, occurCountU = np.unique(U, axis=0, return_counts=True)
    uniqueRowsV, occurCountV = np.unique(V, axis=0, return_counts=True)
    
    D = EuclideanDistance(uniqueRowsU, uniqueRowsV)**p
    n, m = len(uniqueRowsU), len(uniqueRowsV)
  
    A = SparseConstraint(n,m)
    b = np.concatenate([occurCountU/sum(occurCountU), occurCountV/sum(occurCountV)])
    c = D.reshape(-1)

    return linprog(-b, A.T, c, bounds=[None, None], method='highs')
    
def WassersteinDistanceMatrix(dataset, timer=True):
    logs = Timer('WSDM', output=False)  
    if isinstance(dataset.index , pd.MultiIndex):
        dataset.index = dataset.index.get_level_values(0)
        
    unit_ids = dataset.index.unique()
    N = len(unit_ids )
    K = np.zeros((N,N))
    
    k = 0
    for i in range(N):
        for j in range(i+1, N):
            opt_res = PairwiseWassersteinDistance(dataset.loc[unit_ids[i]], 
                                                  dataset.loc[unit_ids[j]])
            K[i,j] = -opt_res.fun
            
            if k%250 == 0 and timer:
                logs.add(f'Completed {k} of {N*(N-1)/2}')
            k+=1
    
    K = np.sqrt(K + K.T)
    
    return pd.DataFrame(K, index=unit_ids , columns=unit_ids)


class GaussianWassersteinDistance:
    def __init__(self, Gaussians, fast_approx=False):
        sqrts = []
        means = []
        covs  = []
        
        self.index = Gaussians.index
        for G in Gaussians:            
            sqrts.append(G.cov.sqrt())
            covs.append(G.cov)
            means.append(G.mean)
            
        self.EDM = self.EuclideanDistanceMatrix(np.stack(means))
        
        if fast_approx:
            self.CDM = self.FrobeniusDistanceMatrix(np.stack(sqrts))
        else:
            self.CDM = self.CovarianceDistanceLoop(covs)
         
    def EuclideanDistanceMatrix(self, X):
        norms  = np.linalg.norm(X, axis=1, ord=2).reshape((len(X),1))**2
        matrix = norms + norms.T - 2 * X@X.T
        return matrix - matrix.min()
    
    def FrobeniusDistanceMatrix(self, X):
        norms = np.linalg.norm(X, ord='fro', axis=(1, 2)).reshape((len(X),1))**2
        matrix = norms + norms.T - 2 * np.tensordot(X,X, axes=([1,2],[1,2]))
        return matrix - matrix.min()
    
    def PairwiseCovarianceDistance(self, cov1, cov2):
        tmp = cov2.sqrt() @ cov1.array() @ cov2.sqrt()
        tmp = arr2cov(tmp)
        tmp = cov1.array() + cov2.array() - 2 * tmp.sqrt()
        return np.sum(np.diag(tmp))

    def CovarianceDistanceLoop(self, covs):
        N = len(covs)
        K = np.zeros((N,N))
        for i in range(N):
            for j in range(i+1,N):
                K[i,j] = self.PairwiseCovarianceDistance(covs[i], covs[j])
        return K + K.T
    
    def matrix(self, w=0.5):
        K =  np.sqrt(2-4*(w-0.5)**2)*np.sqrt((1-w)*self.EDM + w*self.CDM)
        return pd.DataFrame(K, index=self.index , columns=self.index)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 14 10:36:56 2021

@author: fsvbach
"""
import numpy as np
import pandas as pd

from .utils import RandomGenerator, GaussianDistribution, WishartDistribution, CovarianceMatrix, RotationMatrix

class HierarchicalGaussianMixture:       
    config = {'units':100, 
              'samples' :30, 
              'features':2,
              'classes':4,
              'ClassMeanDistance': 20,
              'ClassScaleVariance': 5}
         
    def __init__(self, seed=None, random=True, **kwargs):
        self.config.update(kwargs)
        
        self.N = self.config['units']
        self.M = self.config['samples']
        self.F = self.config['features']
        self.K = self.config['classes']
        
        self.a = self.config['ClassMeanDistance']
        self.b = self.config['ClassScaleVariance']
        
        self.seed = seed
        self.generator = RandomGenerator(seed)
                
        if random:
            self.info = self._info()
            self.set_params()
            self.generate_data()
        else:
            self.info = f'Custom parameters'
            self.ClassGaussians = None
            self.ClassWisharts  = None
        
        self.info += f"\n{self.K} classes à {self.N} datapoints with each {self.M} samples in {self.F} dimensions"
                    
    def _info(self):
        return f'''Random seed: {self.seed}, ClassMeanDistance: {self.a}, ClassScaleVariance: {self.b}'''
    
    def set_params(self, means=None, Lambdas=None, nus=None, Gammas=None):
        prior = lambda b: WishartDistribution(self.F, CovarianceMatrix(np.eye(self.F), b*np.ones(self.F)))
        
        try:
            assert means.shape == (self.K, self.F)
        except:
            print('Sampling means')
            means = self.generator.UniformVector(self.F, self.a, self.K)            
        
        try:
            assert nus.shape == (self.K,)
        except:
            print('Sampling nus')
            nus = self.generator.UniformInteger(lower=self.F, upper=self.F*2, size=self.K)
        
        try:
            assert Gammas[0].array().shape == (self.F, self.F)
        except:
            print('Sampling Gammas')
            Gammas = self.generator.WishartSamples(prior(self.b), self.K)

        try:
            assert Lambdas[0].array().shape == (self.F, self.F)
        except:
            print('Sampling Lambdas')
            Lambdas = self.generator.WishartSamples(prior(1), self.K)
        
        self.ClassGaussians = [GaussianDistribution(mean, Cov) for mean, Cov in zip(means, Gammas)]
        self.ClassWisharts  = [WishartDistribution(nu, Scale) for nu, Scale in zip(nus, Lambdas)]           
        
    def generate_data(self): 
        assert self.ClassGaussians and self.ClassWisharts  
        
        dataset = []
        
        for Gaussian, Wishart in zip(self.ClassGaussians, self.ClassWisharts):

            data_means = self.generator.GaussianSamples(Gaussian, self.N)
            data_covs  = self.generator.WishartSamples(Wishart, self.N)
            
            for mean, cov in zip(data_means, data_covs):
                datapoint = GaussianDistribution(mean, cov)
                dataset.append(self.generator.GaussianSamples(datapoint, self.M))

        index       = pd.MultiIndex.from_product([range(self.K*self.N), range(self.M)], 
                                                 names=["Unit", "Sample"])
        self.data   = pd.DataFrame(np.vstack(dataset), index = index)
        return self.data
    
    def labeldict(self):
        return {i: 'C'+str(i//self.N) for i in range(self.K*self.N)}


def ToyDataset():
    mixture = HierarchicalGaussianMixture(seed=13,
                                        units=100, 
                                        samples=30, 
                                        features=2, 
                                        classes=4,
                                        random=False)

    C = CovarianceMatrix(RotationMatrix(20), s=[10,0.5])
    D = CovarianceMatrix(RotationMatrix(110), s=[10,0.5])
    
    mixture.set_params(means   = np.array([[30,0],[30,0],[0,0],[0,0]]),
                       Gammas = [CovarianceMatrix(s=[5,5])]*4,
                       nus     = np.ones(4)*4,
                       Lambdas  = [C,D,C,D])
    
    return mixture.generate_data(), mixture




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May  7 20:18:09 2022

@author: bachmann
"""

from sklearn.metrics import accuracy_score, cluster
from sklearn.neighbors import KNeighborsClassifier, kneighbors_graph
import igraph as ig 
import leidenalg as la

def knnAccuracy(embedding, labeldict, k=5):
    """
    Input:
    embedding: pd.DataFrame
    labeldict: dictionary that maps the index  of embedding to labels
    
    Returns
    -------
    Accuracy as float.
    
    Note: Since it is not wanted that the label of the test point is used for classification, we multiply the resulting accuracy by the factor of (k+1)/k and subtract 1/k.
    """
    labels = embedding.index.to_series().map(labeldict)
    kNN    = KNeighborsClassifier(k+1)
    kNN.fit(embedding.values, labels)
    estmte = kNN.predict(embedding.values)
    return accuracy_score(estmte, labels)*(k+1)/k-1/k, estate

def LeidenClusters(D, labeldict, k=5, res = 0.08, seed=None):
    kNNgraph = kneighbors_graph(D, k)
    sources, targets = kNNgraph.nonzero()
    G = ig.Graph(directed=True)
    G.add_vertices(kNNgraph.shape[0])
    G.add_edges(list(zip(sources, targets)))
    clustering = la.find_partition(G, la.RBConfigurationVertexPartition, 
                                      resolution_parameter = res, seed=seed)
    labels = D.index.to_series().map(labeldict)
    return cluster.adjusted_rand_score(clustering.membership, labels), clustering.membership